#!/usr/bin/env ruby

require 'erb'
require 'json'
require 'optparse'
require 'yaml'

main_interface = case ENV['PROVIDER']
  when 'hetzner'
    'br0'
  else
    'eth0'
end

options = {
  :encrypt        => true,
  :num_interfaces => 1,
  :partition      => false,
}
optparse = OptionParser.new do |opts|
  opts.banner = "Usage: #{File.basename($0)} [options]"

  opts.on('-d', '--domain DOMAIN', 'DNS domain') { |v| options[:domain] = v }
  opts.on('-D', '--disk DISK', 'VM disk') { |v| options[:disk] = v }
  opts.on('-f', '--facts JSON', 'Puppet facts (JSON string)') { |v| options[:facts] = v }
  opts.on('-i', '--interfaces NUM', 'Number of interfaces the VM will have') { |v| options[:num_interfaces] = v.to_i }
  opts.on('-m', '--mtu MTU', 'MTU for bridge interface(s)') { |v| options[:mtu] = v }
  opts.on('-n', '--nameserver NAMESERVER', 'DNS nameservers (comma-separated') { |v| options[:nameserver] = v }
  opts.on('-p', '--partition', 'Partition the VM disk') { options[:partition] = true }
  opts.on('-s', '--ssh-key KEY', 'Operator SSH public key to be deployed as authorized key (currently only RSA is supported)') { |v| options[:ssh_key] = v }
  opts.on('-x', '--no-encrypt', 'Disable encryption of VM data volume') { options[:encrypt] = false }
end
optparse.parse!

facts = JSON.parse(options[:facts])

# AEP minions must be configured as container-hyervisors first, b/c they need
# some additional data from the Trick operator before they can take on their
# actual role. Hence we need to hide that role during the initial Puppet runs.
actual_role      = facts['v1_role']
facts['v1_role'] = 'container-hypervisor'

raise OptionParser::MissingArgument.new(optparse) if options[:disk].nil?
raise OptionParser::MissingArgument.new(optparse) if options[:domain].nil?
raise OptionParser::MissingArgument.new(optparse) if options[:nameserver].nil?
raise OptionParser::InvalidArgument.new(optparse) if options[:num_interfaces] < 1
if ['hetzner'].include? ENV['PROVIDER']
  vxlan_id = (actual_role == 'aep-minion') ? 20 : 10
  # Overlay interfaces need a fixed MAC address, otherwise the bridge interface
  # would get a new MAC address (and thus a new IP from DHCP) every time the
  # interface comes up, which among other things would cause issues with name
  # resolution (delay until DNS is updated and the update propagated, original
  # IP still in hosts file, ...).
  #
  # Xe:XX:XX are OUIs for Locally Administered Addresses, so that we won't have
  # address conflicts either with NICs from other vendors or between management
  # servers and AEP minions.
  oui      = (actual_role == 'aep-minion') ? 'be:ca:fe' : 'ae:ca:fe'
end

if options[:partition]
  data_volume = "#{options[:disk]}2"
else
  data_volume = options[:disk]
end

if options[:encrypt]
  zpool_member = "#{data_volume.sub(/^\/dev\//, '\&mapper/')}_crypt"
else
  zpool_member = data_volume
end

def load_template(name)
  provider_template = "#{ENV['LIBDIR']}/templates/#{ENV['PROVIDER']}/#{name}.erb"
  default_template  = "#{ENV['LIBDIR']}/templates/default/#{name}.erb"
  template = File.file?(provider_template) ? provider_template : default_template
  return ERB.new(File.read(template))
end

zfs_fact = load_template('zfs_fact').result(binding)

puppet_script          = load_template('puppet_script').result(binding)
datadisk_script        = load_template('datadisk_script').result(binding)
partition_script       = load_template('partition_script').result(binding)
disk_encryption_script = load_template('disk_encryption_script').result(binding)

# Addtitional network interfaces are down after VM creation. Create a bridge on
# each of them and bring it online.
# loop to wait for additional interfaces to appear doesn't need a timeout
# because we already have a timeout on the entire VM creation process.
network_setup_script = load_template('network_setup_script').result()

resolv_conf = <<-EOS
search #{options[:domain]}
options timeout:1 attempts:3
EOS
options[:nameserver].split(',').each do |ns|
resolv_conf << "nameserver #{ns}\n"
end

def new_file(path, content, permissions='0644', owner='root:root')
  {
    'path'        => path,
    'permissions' => permissions,
    'owner'       => owner,
    'content'     => content,
  }
end

# Must prevent cloud-init from populating sources.list b/c requests to public
# apt servers might not work, which would unduly delay deployments (and thus
# might cause the deployment to run into a timeout).
cloud_config = {
  'apt' => {
    'preserve_sources_list' => false,
    'sources_list'          => '# Repos managed by puppet.',
  }
}
cloud_config['resize_rootfs'] = false if options[:partition]
cloud_config['write_files'] = [
  new_file('/etc/resolv.conf', resolv_conf),
  new_file('/etc/facter/facts.d/asag.yaml', YAML.dump(facts)),
  new_file('/etc/facter/facts.d/zfs.yaml', zfs_fact),
  new_file('/usr/local/sbin/run_puppet.sh', puppet_script, '0750')
]
cloud_config['write_files'] << new_file('/usr/local/sbin/prepare_datadisk.sh', datadisk_script, '0750')
cloud_config['write_files'] << new_file('/usr/local/sbin/fix_partitions.sh', partition_script, '0750') if options[:partition]
cloud_config['write_files'] << new_file('/usr/local/sbin/enable_disk_encryption.sh', disk_encryption_script, '0750') if options[:encrypt]
cloud_config['write_files'] << new_file('/usr/local/sbin/network_setup.sh', network_setup_script, '0750')
cloud_config['write_files'] << new_file('/root/.ssh/auth.operator', "ssh-rsa #{options[:ssh_key]}", '0600') unless options[:ssh_key].nil?
cloud_config['runcmd'] = [
  '/usr/local/sbin/run_puppet.sh',
]
cloud_config['runcmd'].unshift '/usr/local/sbin/network_setup.sh'
cloud_config['runcmd'].unshift '/usr/local/sbin/enable_disk_encryption.sh' if options[:encrypt]
cloud_config['runcmd'].unshift '/usr/local/sbin/prepare_datadisk.sh'
cloud_config['runcmd'].unshift '/usr/local/sbin/fix_partitions.sh' if options[:partition]

puts '#cloud-config'
puts cloud_config.to_yaml(options = {:line_width => -1, :use_block => true})
