#!/usr/bin/env ruby

require 'json'
require 'optparse'
require 'yaml'

remote = 'aep'

options = {
  :environment => 'production',
  :facts       => '{"v1_role":"base"}',
  :image       => 'base',
  :immediate   => false,
  :privileged  => false,
}
optparse = OptionParser.new do |opts|
  opts.banner = "Usage: #{File.basename($0)} [options]"

  opts.on('-d', '--domain DOMAIN', 'DNS domain') { |v| options[:domain] = v }
  opts.on('-e', '--environment ENVIRONMENT', 'Puppet environment') { |v| options[:environment] = v }
  opts.on('-f', '--facts JSON', 'External facts for the server') { |v| options[:facts] = v }
  opts.on('-i', '--image IMAGE', "Container image name (the prefix '#{remote}/' is added automatically)") { |v| options[:image] = v }
  opts.on('-I', '--immediate', "Don't postpone role setup") { options[:immediate] = true }
  opts.on('-m', '--puppetmaster SERVER', 'Puppetmaster FQDN or IP address') { |v| options[:puppetmaster] = v }
  opts.on('-n', '--nameserver NAMESERVER', 'DNS nameservers (comma-separated)') { |v| options[:nameserver] = v }
  opts.on('-p', '--pod NAME', 'Pod name') { |v| options[:pod] = v }
  opts.on('-P', '--privileged', 'Create a privileged container') { options[:privileged] = true }
  opts.on('-s', '--ssh-key KEY', 'Operator SSH public key to be deployed as authorized key (currently only RSA is supported)') { |v| options[:ssh_key] = v }
end
optparse.parse!

raise OptionParser::MissingArgument.new(optparse) if options[:domain].nil?
raise OptionParser::MissingArgument.new(optparse) if options[:nameserver].nil?
raise OptionParser::MissingArgument.new(optparse) if options[:puppetmaster].nil?
raise OptionParser::MissingArgument.new(optparse) if options[:pod].nil?

fqdn = "#{options[:pod]}.#{options[:domain]}"
podinfo_path = '/var/lib/aep/podinfo'

facts = JSON.parse(options[:facts])

role = facts['v1_role']

dns_alt_names = [ fqdn ]
dns_alt_names << "trick-operator.#{options[:domain]}" if facts['v1_role'] == 'trick-operator'

puppet_script = <<-EOS
#!/bin/bash
export LANG='C.UTF-8'
ADDR="$(ip -br a s eth0 | awk '{sub(/\\/.*/,"",$3); print $3}')"
if [ -z "$ADDR" ]; then
  echo 'Cannot find eth0 IP address.' 1>&2
  exit 1
fi
cat <<EOF >/etc/hosts
# /etc/hosts
127.0.0.1 localhost
${ADDR} #{fqdn} #{options[:pod]}
178.250.132.131 apt-1.asag.io apt-1
# The following lines are desirable for IPv6 capable hosts
::1 ip6-localhost ip6-loopback
fe00::0 ip6-localnet
ff00::0 ip6-mcastprefix
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
ff02::3 ip6-allhosts
EOF
set +e
for i in {1..10}; do
  rc='0'
  apt-get -q -y update >/dev/null
  /opt/puppetlabs/bin/puppet agent --enable
  /usr/bin/timeout -k 30 300  /opt/puppetlabs/bin/puppet agent --test && break
  rc="${?:-1}"
  /opt/puppetlabs/bin/puppet agent --disable 'initial run'
  [ "$rc" -ne '1' ] || sleep 30  # another puppet run already in progress
done
[ "$rc" -eq '0' ] || exit "$rc"
EOS

unless options[:immediate]
  # pods should get only the base config during the initial Puppet run
  # fact v1_role is changed back to the correct role afterwards
  facts['v1_role'] = 'base'

  puppet_script += <<-EOS
rm -f /etc/cron.d/puppet
sed -i -e 's/\\(v1_role:\\s*\\).*/\\1"#{role}"/' /etc/facter/facts.d/asag.yaml && /opt/puppetlabs/bin/puppet agent --test --noop || true
EOS
end

# register well-known CNAME for trick-operator
if role == 'trick-operator'
  puppet_script += <<-EOS
. /opt/aep/etc/dns_credentials
curl -X PATCH "${DNS_API}/zones/#{options[:domain]}." -ksSL -u "${USERNAME}:${PASSWORD}" --data-binary @- <<EOF
{
  "rrsets": [{
    "name": "#{role}.#{options[:domain]}.",
    "type": "CNAME",
    "ttl": 300,
    "changetype": "REPLACE",
    "records": [{
      "content": "#{fqdn}.",
      "disabled": false,
      "set-ptr": false
    }]
  }]
}
EOF
EOS
end

puppet_config = <<-EOS
[main]
  environment = #{options[:environment]}
[agent]
  server = #{options[:puppetmaster]}
  ca_server = #{options[:puppetmaster]}
  environment = #{options[:environment]}
  dns_alt_names = #{dns_alt_names.join(',')}
EOS

resolv_conf = <<-EOS
search #{options[:domain]}
EOS
options[:nameserver].split(',').each do |ns|
resolv_conf << "nameserver #{ns}\n"
end

cloud_config = {
  'hostname' => options[:pod],
  'fqdn' => fqdn,
  'write_files'    => [{
    'path'        => '/etc/facter/facts.d/asag.yaml',
    'permissions' => '0644',
    'owner'       => 'root:root',
    'content'     => facts.to_yaml.sub(/^---\n/, '').strip,
  },{
    'path'        => '/etc/resolv.conf',
    'permissions' => '0755',
    'owner'       => 'root:root',
    'content'     => resolv_conf,
  },{
    'path'        => '/etc/puppetlabs/puppet/puppet.conf',
    'permissions' => '0750',
    'owner'       => 'root:root',
    'content'     => puppet_config,
  },{
    'path'        => '/usr/local/sbin/run_puppet.sh',
    'permissions' => '0750',
    'owner'       => 'root:root',
    'content'     => puppet_script,
  },{
    'path'        => '/root/.ssh/auth.operator',
    'permissions' => '0600',
    'owner'       => 'root:root',
    'content'     => "ssh-rsa #{options[:ssh_key]} operator-on-root",
  }],
  'runcmd' => [
    'ssh-keyscan gitlab-2.asag.io >> /root/.ssh/known_hosts',
    'bash /usr/local/sbin/run_puppet.sh',
  ],
}
user_data = <<-EOS
#cloud-config
#{cloud_config.to_yaml.sub(/^---\n/, '')}
EOS

manifest = {
  'apiVersion' => 'v1',
  'kind'       => 'Pod',
  'metadata' => {
    'name' => options[:pod],
    'labels' => {
      'dns.aep.asag.io/fqdn' => fqdn,
      'dns.aep.asag.io/name' => options[:pod],
    },
  },
  'spec' => {
    'hostname'   => options[:pod],
    'containers' => [{
      'name'            => options[:pod],
      'image'           => "#{remote}/#{options[:image]}",
      'imagePullPolicy' => 'Always',
      'securityContext' => {
        'privileged' => options[:privileged],
      },
      'env' => [{
        'name'  => 'user-data',
        'value' => user_data,
      }],
      'volumeMounts' => [{
        'mountPath' => podinfo_path,
        'name'      => 'aep-podinfo',
      }],
    }],
    'volumes' => [{
      'downwardAPI' => {
        'items' => [{
          'fieldRef' => { 'apiVersion' => 'v1', 'fieldPath' => 'metadata.labels' },
          'path'     => 'labels',
        }, {
          'fieldRef' => { 'apiVersion' => 'v1', 'fieldPath' => 'metadata.annotations' },
          'path'     => 'annotations',
        }, {
          'fieldRef' => { 'apiVersion' => 'v1', 'fieldPath' => 'metadata.name' },
          'path'     => 'name',
        }, {
          'fieldRef' => { 'apiVersion' => 'v1', 'fieldPath' => 'metadata.namespace' },
          'path'     => 'namespace',
        }],
      },
      'name' => 'aep-podinfo',
    }],
  },
}

if options[:privileged]
  #NOTE: The additional LXD config must be specified as Hash[String, String]!
  manifest['metadata']['annotations'] = {
    'x-lxe-additional-lxd-config' => {
      'raw.lxc'          => "lxc.mount.auto = cgroup:rw:force\nlxc.mount.auto = proc:rw",
      'security.nesting' => 'true',
    }
  }
end

puts manifest.to_yaml
